/********************************************************************/
/*                                                                  */
/*            L   I  QQ  U U I DD    W   W  A  RR    555            */
/*            L   I Q  Q U U I D D   W   W A A R R   5              */
/*            L   I Q  Q U U I D D   W W W AAA RR    55             */
/*            L   I Q Q  U U I D D   WW WW A A R R     5            */
/*            LLL I  Q Q  U  I DD    W   W A A R R   55             */
/*                                                                  */
/*                             b                                    */
/*                             bb  y y                              */
/*                             b b yyy                              */
/*                             bb    y                              */
/*                                 yy                               */
/*                                                                  */
/*                     U U       FFF  O   O  TTT                    */
/*                     U U       F   O O O O  T                     */
/*                     U U TIRET FF  O O O O  T                     */
/*                     U U       F   O O O O  T                     */
/*                      U        F    O   O   T                     */
/*                                                                  */
/********************************************************************/

/********************************************************************/
/* this software is protected by the GPL, see copying.txt           */
/********************************************************************/

/********************************************************************/
/* name          : srvchan.c                                        */
/* content       : handles a server network channel                 */
/* date de modif : 3 mai 98                                         */
/********************************************************************/

/*==================================================================*/
/* includes                                                         */
/*==================================================================*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "srvchan.h"
#include "netmess.h"
#include "log.h"
#include "netmap.h"
#include "netconf.h"
#include "sock2gen.h"

/*==================================================================*/
/* static functions                                                 */
/*==================================================================*/

static void reset(LW_SRVCHAN *chan);
static int handle_ping(LW_SRVCHAN *chan, LW_NETMESS *mess);
static int handle_echo(LW_SRVCHAN *chan, LW_NETMESS *mess);
static int handle_program(LW_SRVCHAN *chan, LW_NETMESS *mess);
static int handle_version(LW_SRVCHAN *chan, LW_NETMESS *mess);
static int handle_nickname(LW_SRVCHAN *chan, LW_NETMESS *mess);
static int handle_free(LW_SRVCHAN *chan, LW_NETMESS *mess,
		       int free_teams);
static int handle_team(LW_SRVCHAN *chan, LW_NETMESS *mess,
		       int *free_teams);
static int handle_ready(LW_SRVCHAN *chan, LW_NETMESS *mess);
static int handle_who(LW_SRVCHAN *chan, LW_NETMESS *mess);
static int handle_sendmap(LW_SRVCHAN *chan, LW_NETMESS *mess, 
			  void **map);
static int handle_recvmap(LW_SRVCHAN *chan, LW_NETMESS *mess, 
			  void *map);
static int handle_sendconfig(LW_SRVCHAN *chan, LW_NETMESS *mess, 
			     LW_NETCONF *config);
static int handle_recvconfig(LW_SRVCHAN *chan, LW_NETMESS *mess, 
			     LW_NETCONF *config);
static int handle_unknown(LW_SRVCHAN *chan, LW_NETMESS *mess);

/*------------------------------------------------------------------*/
/*
 * Resets the LW_SRVCHAN struct
 */
static void reset(LW_SRVCHAN *chan)
{
  int i;

  chan->sock=-1;
  chan->ip[0]='\0';
  chan->port=0;
  chan->program_ok=1; // set to 1 temporary for testing
  chan->version_ok=1; // set to 1 temporary for testing
  chan->nickname_ok=0;
  chan->ready_ok=0;
  chan->nickname[0]='\0';
  for (i=0;i<NB_TEAMS;++i)
    {
      lw_srvteam_reset(&(chan->team[i]));
    }
}

/*------------------------------------------------------------------*/
/*
 * handles the "PING" message
 */
static int handle_ping(LW_SRVCHAN *chan, LW_NETMESS *mess)
{
  int result=0;

  /*
   * It's a ping, we answer OK everytime
   */
  lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_OK);
  result=1;

  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "ECHO" message
 */
static int handle_echo(LW_SRVCHAN *chan, LW_NETMESS *mess)
{
  int result=0;
  char answer[LW_SOCK_MESSAGE_SIZE];
  int i;

  /*
   * echo, we rebuild the original message
   * from the information we have in the LW_NETMESS
   * struct, and resend it to the client
   */
  strcpy(answer,LW_NETMESS_TEXT_ECHO);
  i=0;
  while (i<mess->argc)
    {
      if (i==0)
	{
	  strcat(answer," ");
	}
      else
	{
	  strcat(answer,",");
	}
      strcat(answer,mess->argv[i]);
      i++;
    }
  lw_sock_send_str(chan->sock,answer);
  result=1;

  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "PROGRAM" message
 */
static int handle_program(LW_SRVCHAN *chan, LW_NETMESS *mess)
{
  int result=0;

  if (mess->argc==1)
    {
      /*
       * OK, there's an argument, we check if it's the good prog name.
       * This prevents form accepting random connections: if the client
       * is not really a liquidwar client, it will be thrown away
       */
      if (strcmp(mess->argv[0],LW_PROGRAM)==0)
	{
	  chan->program_ok=1;
	  result=1;
	}
    }

  /*
   * Now we send feedback on the network
   */
  if (result)
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_OK);
    }
  else
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_ERR);
    }

  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "VERSION" message
 */
static int handle_version(LW_SRVCHAN *chan, LW_NETMESS *mess)
{
  int result=0;

  if (mess->argc==1)
    {
      /*
       * OK, there's an argument, we check if it's the good version.
       * This prevents form accepting random connections: if the client
       * is not a liquidwar client with the right version number, it will
       * be thrown away...
       */
      if (strcmp(mess->argv[0],LW_VERSION)==0)
	{
	  chan->version_ok=1;
	  result=1;
	}
    }

  /*
   * Now we send feedback on the network
   */
  if (result)
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_OK);
    }
  else
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_ERR);
    }

  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "NICKNAME" message
 */
static int handle_nickname(LW_SRVCHAN *chan, LW_NETMESS *mess)
{
  int result=0;

  if (mess->argc==1)
    {
      /*
       * OK, there's an argument, we copy it to the nickname field.
       */
      strncpy(chan->nickname,mess->argv[0],sizeof(chan->nickname)-1);
      chan->nickname[sizeof(chan->nickname)-1]='\0';
      chan->nickname_ok=1;
      result=1;
    }

  /*
   * Now we send feedback on the network
   */
  if (result)
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_OK);
    }
  else
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_ERR);
    }

  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "FREE" message
 */
static int handle_free(LW_SRVCHAN *chan, LW_NETMESS *mess,int free_teams)
{
  int result=0;
  char answer[LW_SOCK_MESSAGE_SIZE];

  sprintf(answer,"%s %d",LW_NETMESS_TEXT_OK,free_teams);

  lw_sock_send_str(chan->sock,answer);
  result=1;

  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "TEAM" message
 */
static int handle_team(LW_SRVCHAN *chan, LW_NETMESS *mess, 
		       int *free_teams)
{
  int result=0;
  int i;

  if (mess->argc==1)
    {
      /*
       * OK, there's an argument, we interpret it as a number
       */
      i=atoi(mess->argv[0]);
      if (i>=0 && i<NB_TEAMS)
	{
	  if (!chan->team[i].active)
	    {
	      /*
	       * The message from the client makes sense, we
	       * return OK
	       */
	      result=1;

	      /*
	       * Now we activate the team only if there's still
	       * room. If not we still return OK but the "WHO"
	       * message will later inform the client that some
	       * of his requests have been cancelled...
	       */
	      if ((*free_teams)>0)
		{
		  (*free_teams)--;
		  chan->team[i].active=1;
		}
	    }
	}
    }

  /*
   * Now we send feedback on the network
   */
  if (result)
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_OK);
    }
  else
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_ERR);
    }

  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "WHO" message
 */
static int handle_who(LW_SRVCHAN *chan, LW_NETMESS *mess)
{
  int result=0;
  int i;
  char answer[LW_SOCK_MESSAGE_SIZE];

  if (mess->argc==1)
    {
      /*
       * OK, there's an argument, we interpret it as a number
       */
      i=atoi(mess->argv[0]);
      if (i>=0 && i<NB_TEAMS)
	{
	  sprintf(answer,"%s",LW_NETMESS_TEXT_ERR);

	  if (chan->team[i].active)
	    {
	      if (chan->team[i].network)
		{
		  sprintf(answer,"%s %d,%s",
			  LW_NETMESS_TEXT_NETWORK,
			  chan->team[i].server_id,
			  chan->team[i].nickname);
		}
	      else
		{
		  sprintf(answer,"%s %d,%s",
			  LW_NETMESS_TEXT_LOCAL,
			  chan->team[i].server_id,
			  chan->team[i].nickname);
		}
	    }
	  else
	    {
	      sprintf(answer,"%s",LW_NETMESS_TEXT_NOBODY);
	    }

	  lw_sock_send_str(chan->sock,answer);
	  result=1;
	}
    }

  /*
   * Now we send feedback on the network
   */
  if (!result)
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_ERR);
    }

  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "READY" message
 */
static int handle_ready(LW_SRVCHAN *chan, LW_NETMESS *mess)
{
  int result=0;

  /*
   * Here we do not return "OK" yet. In fact we will return
   * "OK" on the network later, when all the clients will be
   * connected, this way the "OK" answer to the "READY" message
   * is the signal for game start
   */
  chan->ready_ok=1;
  result=1;

  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "SENDMAP" message
 */
static int handle_sendmap(LW_SRVCHAN *chan, 
			  LW_NETMESS *mess, 
			  void **map)
{
  int result=0;

  result=lw_netmap_recv(chan->sock,map);

  /*
   * Now we send feedback on the network
   */
  if (result)
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_OK);
    }
  else
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_ERR);
    }


  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "RECVMAP" message
 */
static int handle_recvmap(LW_SRVCHAN *chan, 
			  LW_NETMESS *mess,
			  void *map)
{
  int result=0;

  result=lw_netmap_send(chan->sock,map);

  /*
   * Now we send feedback on the network
   */
  if (result)
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_OK);
    }
  else
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_ERR);
    }

  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "SENDCONFIG" message
 */
static int handle_sendconfig(LW_SRVCHAN *chan,
			     LW_NETMESS *mess,
			     LW_NETCONF *config)
{
  int result=0;

  result=lw_netconf_recv(chan->sock,config);

  /*
   * Now we send feedback on the network
   */
  if (result)
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_OK);
    }
  else
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_ERR);
    }


  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles the "RECVCONFIG" message
 */
static int handle_recvconfig(LW_SRVCHAN *chan,
			     LW_NETMESS *mess,
			     LW_NETCONF *config)
{
  int result=0;

  result=lw_netconf_send(chan->sock,config);

  /*
   * Now we send feedback on the network
   */
  if (result)
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_OK);
    }
  else
    {
      lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_ERR);
    }

  return result;
}

/*------------------------------------------------------------------*/
/*
 * handles an unknown message
 */
static int handle_unknown(LW_SRVCHAN *chan, LW_NETMESS *mess)
{
  int result=0;

  /*
   * We answer ERR all the time
   */
  lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_ERR);

  /*
   * Since we return 0, the connection will be closed...
   */
  return result;
}

/*==================================================================*/
/* global functions                                                 */
/*==================================================================*/

/*------------------------------------------------------------------*/
/*
 * Waits for a client to connect, returns 1 if successfull, 0 if not
 */
int lw_srvchan_wait_teams(LW_SRVCHAN *chan, int *free_teams,
			  int sock,void **map, LW_NETCONF *config)
{
  int result=0;
  int ret;
  int ret_mess=1;
  char message[LW_SOCK_MESSAGE_SIZE];
  LW_NETMESS *mess;

  reset(chan);

  log_print_str("Waiting for ");
  log_print_int(*free_teams);
  if ((*free_teams)>1)
    {
      log_print_str(" teams...");
    }
  else
    {
      log_print_str(" team...");
    }
  log_println();
  if (lw_sock_accept(&(chan->sock),chan->ip,&(chan->port),sock))
    {
      log_print_str("Connection from ");
      log_print_str(chan->ip);
      log_print_str(":");
      log_print_int(chan->port);
      log_println();
      log_flush();

      while ((!chan->ready_ok) && ret_mess &&
	     ((ret=lw_sock_recv_str(chan->sock,message))>=0))
	{
	  if (ret>0)
	    {
	      mess=lw_netmess_read(message);
	      if (mess!=NULL)
		{
		  /*
		   * switch to handle the different possible messages
		   */
		  switch (mess->code)
		    {
		    case LW_NETMESS_CODE_PING:
		      ret_mess=handle_ping(chan,mess);
		      break;
		    case LW_NETMESS_CODE_ECHO:
		      ret_mess=handle_echo(chan,mess);
		      break;
		    case LW_NETMESS_CODE_PROGRAM:
		      ret_mess=handle_program(chan,mess);
		      break;
		    case LW_NETMESS_CODE_VERSION:
		      ret_mess=handle_version(chan,mess);
		      break;
		    case LW_NETMESS_CODE_NICKNAME:
		      ret_mess=handle_nickname(chan,mess);
		      break;
		    case LW_NETMESS_CODE_FREE:
		      ret_mess=handle_free(chan,mess,*free_teams);
		      break;
		    case LW_NETMESS_CODE_TEAM:
		      ret_mess=handle_team(chan,mess,free_teams);
		      break;
		    case LW_NETMESS_CODE_READY:
		      ret_mess=handle_ready(chan,mess);
		      break;
		    case LW_NETMESS_CODE_SENDMAP:
		      ret_mess=handle_sendmap(chan,mess,map);
		      break;
		    case LW_NETMESS_CODE_SENDCONFIG:
		      ret_mess=handle_sendconfig(chan,mess,config);
		      break;
		    default:
		      ret_mess=handle_unknown(chan,mess);
		      break;
		    }

		  /*
		   * lw_netmess_free must *not* be forgetted 
		   */
		  lw_netmess_free(mess);

		  if (!ret_mess)
		    {
		      log_print_str("Error processing \"");
		      log_print_str(message);
		      log_print_str("\"");
		      log_println();
		    }
		}
	      else
		{
		  log_println_str("Serious memory problem 8-(");
		  ret_mess=0;
		}
	    }
	}

      if (chan->program_ok && 
	  chan->version_ok && 
	  chan->nickname_ok &&
	  chan->ready_ok &&
	  ret_mess)
	{
	  result=1;
	  log_print_str("Accepted \"");
	  log_print_str(chan->nickname);
	  log_print_str("\"");
	  log_println();
	}
      else
	{
	  if (chan->ready_ok)
	    {
	      /*
	       * Now if we are here it means we have received a "READY"
	       * message, but still the client did not send the correct
	       * messages before => we send an "ERR" messageon the
	       * network for consistency, since the client did not receive
	       * any message to his "READY" query yet.
	       */
	      lw_sock_send_str(chan->sock, LW_NETMESS_TEXT_ERR);
	    }
	  lw_sock_close(chan->sock);
	  chan->sock=-1;
	  log_println_str("Rejected!");
	}
    }

  return result;
} 

/*------------------------------------------------------------------*/
/*
 * Answers OK to a client, and waits for the "WHO" message 
 */
int lw_srvchan_tell_who(LW_SRVCHAN *chan,void *map,LW_NETCONF *config)
{
  int result=0;
  int ret;
  int ret_mess=1;
  char message[LW_SOCK_MESSAGE_SIZE];
  LW_NETMESS *mess;

  log_print_str("Talking to \"");
  log_print_str(chan->nickname);
  log_print_str("\"");
  log_println();

  /*
   * we set ready_ok back to false since we will wait for
   * another ready message
   */
  chan->ready_ok=0;

  /*
   * We first answer the "READY" message which had been sent by the
   * client a possibly long time ago.
   */
  if (lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_OK))
    {
      while ((!chan->ready_ok) && ret_mess &&
	     ((ret=lw_sock_recv_str(chan->sock,message))>=0))
	{
	  if (ret>0)
	    {
	      mess=lw_netmess_read(message);
	      if (mess!=NULL)
		{
		  /*
		   * switch to handle the different possible messages
		   */
		  switch (mess->code)
		    {
		    case LW_NETMESS_CODE_PING:
		      ret_mess=handle_ping(chan,mess);
		      break;
		    case LW_NETMESS_CODE_ECHO:
		      ret_mess=handle_echo(chan,mess);
		      break;
		    case LW_NETMESS_CODE_WHO:
		      ret_mess=handle_who(chan,mess);
		      break;
		    case LW_NETMESS_CODE_READY:
		      ret_mess=handle_ready(chan,mess);
		      break;
		    case LW_NETMESS_CODE_RECVMAP:
		      ret_mess=handle_recvmap(chan,mess,map);
		      break;
		    case LW_NETMESS_CODE_RECVCONFIG:
		      ret_mess=handle_recvconfig(chan,mess,config);
		      break;
		    default:
		      ret_mess=handle_unknown(chan,mess);
		      break;
		    }

		  /*
		   * lw_netmess_free must *not* be forgetted 
		   */
		  lw_netmess_free(mess);

		  if (!ret_mess)
		    {
		      log_print_str("Error processing \"");
		      log_print_str(message);
		      log_print_str("\"");
		      log_println();
		    }
		}
	      else
		{
		  log_println_str("Serious memory problem 8-(");
		  ret_mess=0;
		}
	    }
	}

      if (chan->ready_ok &&
	  ret_mess)
	{
	  result=1;
	  //	  log_println_str("Player OK");
	}
      else
	{
	  if (chan->ready_ok)
	    {
	      /*
	       * Now if we are here it means we have received a "READY"
	       * message, but still the client did not send the correct
	       * messages before => we send an "ERR" messageon the
	       * network for consistency, since the client did not receive
	       * any message to his "READY" query yet.
	       */
	      lw_sock_send_str(chan->sock, LW_NETMESS_TEXT_ERR);
	    }
	  lw_sock_close(chan->sock);
	  chan->sock=-1;
	  //log_println_str("Player KO");
	}
    }

  return result;
} 

/*------------------------------------------------------------------*/
/*
 * Sends the final "OK" message before streaming
 */
int lw_srvchan_final_ok(LW_SRVCHAN *chan)
{
  int result=0;

  if (lw_sock_send_str(chan->sock,LW_NETMESS_TEXT_OK))
    {
      result=1;
    }
  else
    {
      lw_sock_close(chan->sock);
      chan->sock=-1;
    }

  return result;
}

/*------------------------------------------------------------------*/
/*
 * Returns the index of the team associated to a given server team id
 */
int lw_srvchan_find_team_by_server_id(LW_SRVCHAN *chan,int server_id)
{
  int result=-1;
  int i;

  for (i=0;i<NB_TEAMS && result<0;++i)
    {
      if (chan->team[i].active && chan->team[i].server_id==server_id)
	{
	  result=i;
	}
    }

  return result;
}

/*------------------------------------------------------------------*/
/*
 * Returns the index of the first inactive team
 */
int lw_srvchan_find_first_free_team(LW_SRVCHAN *chan)
{
  int result=-1;
  int i;

  for (i=0;i<NB_TEAMS && result<0;++i)
    {
      if (!(chan->team[i].active))
	{
	  result=i;
	}
    }

  return result;
}

/*------------------------------------------------------------------*/
/*
 * Closes everything and frees the resources
 */
void lw_srvchan_close(LW_SRVCHAN *chan)
{
  lw_sock_close(chan->sock);

  reset(chan);
}


/*------------------------------------------------------------------*/
/*
 * Receives key presses from a computer.
 * The keys come in the "computer's order" ie one still has to
 * sort them by server id. 
 */
int lw_srvchan_recv_keys(LW_SRVCHAN *chan, LW_NETKEY *netkey)
{
  int result=0;
  int i;
  LW_NETKEY chan_netkey;

  lw_netkey_reset(&chan_netkey);

  result=lw_netkey_recv(chan->sock,&chan_netkey);

  for (i=0;i<NB_TEAMS;++i)
    {
      if (chan->team[i].active && !chan->team[i].network)
	{
	  netkey->key_states[chan->team[i].server_id]=
	    chan_netkey.key_states[i];
	}
    }

  netkey->cmd=chan_netkey.cmd;
  netkey->arg=chan_netkey.arg;

  return result;
}

/*------------------------------------------------------------------*/
/*
 * Sends key presses to a computer.
 * The keys must be send in the "computer's order" ie they must
 * be sorted first. 
 */
int lw_srvchan_send_keys(LW_SRVCHAN *chan, LW_NETKEY *netkey)
{
  int result=0;
  int i;
  LW_NETKEY chan_netkey;

  lw_netkey_reset(&chan_netkey);

  for (i=0;i<NB_TEAMS;++i)
    {
      if (chan->team[i].active)
	{
	  chan_netkey.key_states[i]=
	    netkey->key_states[chan->team[i].server_id];
	}
    }

  chan_netkey.cmd=netkey->cmd;
  chan_netkey.arg=netkey->arg;

  result=lw_netkey_send(chan->sock,&chan_netkey);

  return result;
}

